Boomy = {}

--
-- VARS
--

Boomy.version = "1.0.3"
Boomy.releaseDate = "Feb 24, 2015"

SLASH_BOOMY1 = "/boomy"

Boomy.initialized = false
Boomy.timeSinceLastUpdate = 0

Boomy.mainFrame = nil
Boomy.spellFrame = nil
Boomy.secondFrame = nil
Boomy.moonfireFrame = nil
Boomy.sunfireFrame = nil
Boomy.starfallFrame = nil
Boomy.starsurgeFrame = nil

Boomy.gcd = 0

Boomy.eclipse = {
	duration = 20,
	power = 0,
	direction = "none",
	startTime = nil,
	remainingTime = 0
}

-- name, icon, castTime, maxRange
Boomy.spells = {
	starfire = {
		id = 2912
	},
	wrath = {
		id = 5176
	},
	starsurge = {
		id = 78674
	},
	starfall = {
		id = 48505,
		duration = 8
	},
	moonfire = {
		id = 8921,
		duration = 40
	},
	sunfire = {
		id = 93402,
		duration = 24
	},
	celestialAlignment = {
		id = 112071
	},
	incarnation = {
		id = 102560,
		cooldown = 180
	}
}

-- name, icon
Boomy.buffs = {
	lunarEmpowerment = {
		id = 164547
	},
	solarEmpowerment = {
		id = 164545
	},
	empoweredMoonkin = {
		id = 157228
	}
}

-- name, icon
-- expirationTime
Boomy.debuffs = {
	moonfire = {
		id = 164812
	},
	sunfire = {
		id = 164815
	}
}

--
-- EVENTS
--

Boomy.events = {}
Boomy.eventFrame =CreateFrame("Frame")

Boomy.eventFrame:SetScript("OnEvent", function(self, event, ...)
	Boomy.events[event](self, ...);
end)

function Boomy.events:ADDON_LOADED(...)
	if Boomy.initialized == false then
		Boomy:Initialize()
	end
end

function Boomy.events:PLAYER_TALENT_UPDATE(...)
	Boomy:UpdateTalents()
end

for k, v in pairs(Boomy.events) do
	Boomy.eventFrame:RegisterEvent(k);
end

Boomy.eventFrame:SetScript("OnUpdate", function(this, elapsed)
	Boomy:OnUpdate(elapsed)
end)

--
-- SLASH COMMANDS
--

SlashCmdList["BOOMY"] = function (msg, editbox)
	local command, rest = msg:match("^(%S*)%s*(.-)$");

	-- lock
	if command == "lock" then
		Boomy:LockFrame(Boomy.mainFrame)
		print("Boomy is locked")

	-- unlock
	elseif command == "unlock" then
		Boomy:UnlockFrame(Boomy.mainFrame)
		print("Boomy is unlocked")

	-- reset
	elseif command == "reset" then
		Boomy:InitDB()
		Boomy:ResetFrame(Boomy.mainFrame)
		print("Boomy reset")

	-- scale
	elseif command == "scale" and rest ~= "" then
		Boomy:Scale(rest)
		print("Boomy scaled to "..rest)

	-- about
	else
		print("Boomy "..Boomy.version.." ("..Boomy.releaseDate..")")
		print("/boomy lock")
		print("/boomy unlock")
		print("/boomy reset")
	end
end

--
-- INITIALIZE
--

function Boomy:Initialize()
	Boomy.initialized = true

	print("Boomy "..Boomy.version.." ("..Boomy.releaseDate..")")
	print("/boomy")

	--
	-- DATABASE
	--

	if not BoomyDB then
		Boomy:InitDB()
	end

	--
	-- SPELLS, BUFFS and DEBUFFS
	--

	for k, spell in pairs(Boomy.spells) do
		spell.name, _, spell.icon, spell.castTime, _, spell.maxRange = GetSpellInfo(spell.id)
	end

	for k, buff in pairs(Boomy.buffs) do
		buff.name, _, buff.icon = GetSpellInfo(buff.id)
	end

	for k, debuff in pairs(Boomy.debuffs) do
		debuff.name, _, debuff.icon = GetSpellInfo(debuff.id)
	end

	--
	-- FRAMES
	--

	-- mainFrame
	local frame = CreateFrame("Frame", nil, UIParent)
	frame:SetFrameStrata("BACKGROUND")
	frame:SetWidth(BoomyDB.frameUnit * 8)
	frame:SetHeight(BoomyDB.frameUnit * 8)
	frame:SetPoint(BoomyDB.rp, BoomyDB.x, BoomyDB.y)
	frame:SetBackdrop({
        bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
        tile = true,
        tileSize = 32
	})
	frame:SetBackdropColor(0, 0, 0, 0);
	frame:SetScript("OnMouseDown", function(self)
		if self:IsMovable() then
			self:StartMoving()
		end
	end)
	frame:SetScript("OnMouseUp", function(self)
		self:StopMovingOrSizing()
		local _, _, rp, x, y = self:GetPoint()
		BoomyDB.rp = rp
		BoomyDB.x = x
		BoomyDB.y = y
	end)
	frame:EnableMouse(false)
	frame:Show()

	Boomy.mainFrame = frame

	-- spellFrame
	Boomy.spellFrame = Boomy:CreateFrame(Boomy.mainFrame, 2, BoomyDB.frames.spell.x, BoomyDB.frames.spell.y)

	-- moonfireFrame
	Boomy.moonfireFrame = Boomy:CreateFrame(Boomy.mainFrame, 1, BoomyDB.frames.moonfire.x, BoomyDB.frames.moonfire.y, Boomy.spells.moonfire.icon)

	-- sunfireFrame
	Boomy.sunfireFrame = Boomy:CreateFrame(Boomy.mainFrame, 1, BoomyDB.frames.sunfire.x, BoomyDB.frames.sunfire.y, Boomy.spells.sunfire.icon)

	-- starfallFrame
	Boomy.starfallFrame = Boomy:CreateFrame(Boomy.mainFrame, 1, BoomyDB.frames.starfall.x, BoomyDB.frames.starfall.y, Boomy.spells.starfall.icon)

	-- starsurgeFrame
	Boomy.starsurgeFrame = Boomy:CreateFrame(Boomy.mainFrame, 1, BoomyDB.frames.starsurge.x, BoomyDB.frames.starsurge.y, Boomy.spells.starsurge.icon)

	-- celestialAlignmentFrame
	Boomy.celestialAlignmentFrame = Boomy:CreateFrame(Boomy.mainFrame, 1, BoomyDB.frames.celestialAlignment.x, BoomyDB.frames.celestialAlignment.y, Boomy.spells.celestialAlignment.icon)

	-- incarnationFrame
	Boomy.incarnationFrame = Boomy:CreateFrame(Boomy.mainFrame, 1, BoomyDB.frames.incarnation.x, BoomyDB.frames.incarnation.y, Boomy.spells.incarnation.icon)

	--
	-- TALENTS
	--

	Boomy:UpdateTalents()
end

--
-- MAIN LOOP
--

function Boomy:OnUpdate(elapsed)
	Boomy.timeSinceLastUpdate = Boomy.timeSinceLastUpdate + elapsed

	if (Boomy.timeSinceLastUpdate > 0.025) then
		Boomy:update()
	end
end

--
-- UPDATE
--

function Boomy:update()
	--
	-- UPDATE VARS
	--

	Boomy.timeSinceLastUpdate = 0;

	-- gcd
	Boomy.gcd = 1.5 / ( 1 + GetRangedHaste() / 100 )

	-- eclipse
	Boomy.eclipse.direction = GetEclipseDirection()

	local power = UnitPower("player", SPELL_POWER_ECLIPSE)
	if (Boomy.eclipse.power <= 0 and power > 0) or (Boomy.eclipse.power > 0 and power <= 0) or Boomy.eclipse.power == 0 then
		Boomy.eclipse.startTime = GetTime()
		Boomy.eclipse.remainingTime = Boomy.eclipse.duration
	else
		Boomy.eclipse.remainingTime = Boomy.eclipse.duration + (Boomy.eclipse.startTime - GetTime())
	end

	Boomy.eclipse.power = power

	-- spells casting time
	local castTime
	for k, spell in pairs(Boomy.spells) do
		if spell.castTime then
			_, _, _, castTime = GetSpellInfo(spell.id)
			castTime = castTime / 1000
			if castTime <= Boomy.gcd then
				castTime = Boomy.gcd
			end
			spell.castTime = castTime
		end
	end

	-- starsurge
	Boomy.spells.starsurge.charges = GetSpellCharges(Boomy.spells.starsurge.id)

	-- starfall
	Boomy.spells.starfall.charges = GetSpellCharges(Boomy.spells.starfall.id)
	Boomy.spells.starfall.expirationTime = Boomy:GetBuff(Boomy.spells.starfall.name)

	-- moonfire
	Boomy.debuffs.moonfire.expirationTime = Boomy:GetDebuff(Boomy.debuffs.moonfire.name)

	-- sunfire
	Boomy.debuffs.sunfire.expirationTime = Boomy:GetDebuff(Boomy.debuffs.sunfire.name)

	-- lunar empowerment
	Boomy.buffs.lunarEmpowerment.expirationTime, Boomy.buffs.lunarEmpowerment.charges = Boomy:GetBuff(Boomy.buffs.lunarEmpowerment.name)

	-- solar empowerment
	Boomy.buffs.solarEmpowerment.expirationTime, Boomy.buffs.solarEmpowerment.charges = Boomy:GetBuff(Boomy.buffs.solarEmpowerment.name)

	-- celestial alignment
	local start
	start, Boomy.spells.celestialAlignment.cooldown = GetSpellCooldown(Boomy.spells.celestialAlignment.id)
	if start ~= 0 then
		Boomy.spells.celestialAlignment.cooldownRemaining = Boomy.spells.celestialAlignment.cooldown - (GetTime() - start)
	else
		Boomy.spells.celestialAlignment.cooldownRemaining = 0
	end

	Boomy.spells.celestialAlignment.expirationTime = Boomy:GetBuff(Boomy.spells.celestialAlignment.name)

	-- empowered moonkin
	Boomy.buffs.empoweredMoonkin.expirationTime = Boomy:GetBuff(Boomy.buffs.empoweredMoonkin.name)

	-- incarnation
	local start = GetSpellCooldown(Boomy.spells.incarnation.id)
	if start ~= 0 then
		Boomy.spells.incarnation.cooldownRemaining = Boomy.spells.incarnation.cooldown - (GetTime() - start)
	else
		Boomy.spells.incarnation.cooldownRemaining = 0
	end

	Boomy.spells.incarnation.expirationTime = Boomy:GetBuff(Boomy.spells.incarnation.name)

	--
	-- UPDATE FRAME
	--

	-- main frame
	local spell = Boomy:GetSpellToCast()

	Boomy:Draw(Boomy.spellFrame, spell.icon)

	if spell.id == Boomy.spells.starfire.id then
		Boomy:SetCounter(Boomy.spellFrame, Boomy.buffs.lunarEmpowerment.charges)
	elseif spell.id == Boomy.spells.wrath.id then
		Boomy:SetCounter(Boomy.spellFrame, Boomy.buffs.solarEmpowerment.charges)
	else
		Boomy:SetCounter(Boomy.spellFrame, 0)
	end

	-- moonfire frame
	Boomy:SetCounter(Boomy.moonfireFrame, math.ceil(Boomy.debuffs.moonfire.expirationTime, 0))
	Boomy:DecreaseFrameAlpha(Boomy.moonfireFrame)

	if (Boomy.eclipse.power <= 0 or Boomy.spells.celestialAlignment.expirationTime > 0) and Boomy.debuffs.moonfire.expirationTime < Boomy.spells.moonfire.duration * 0.3 then -- (lunar eclipse OR celestial alignment) AND moonfire expiration time < moonfire duration * 30%
		Boomy:IncreaseFrameAlpha(Boomy.moonfireFrame)
	end

	-- sunfire frame
	Boomy:SetCounter(Boomy.sunfireFrame, math.ceil(Boomy.debuffs.sunfire.expirationTime, 0))
	Boomy:DecreaseFrameAlpha(Boomy.sunfireFrame)

	if (Boomy.eclipse.power > 0 or Boomy.spells.celestialAlignment.expirationTime > 0) and Boomy.debuffs.sunfire.expirationTime < Boomy.spells.sunfire.duration * 0.3 then -- (solar eclipse OR celestial alignment) AND sunfire expiration time < sunfire duration * 30%
		Boomy:IncreaseFrameAlpha(Boomy.sunfireFrame)
	end

	-- starfall frame
	Boomy:SetCounter(Boomy.starfallFrame, math.ceil(Boomy.spells.starfall.expirationTime, 0))
	Boomy:DecreaseFrameAlpha(Boomy.starfallFrame)

	if Boomy.spells.starfall.charges > 0 and Boomy.spells.starfall.expirationTime == 0 then -- starfall charge > 0 AND no starfall running
		Boomy:IncreaseFrameAlpha(Boomy.starfallFrame)
	end

	-- starsurge frame
	Boomy:SetCounter(Boomy.starsurgeFrame, Boomy.spells.starsurge.charges)
	Boomy:DecreaseFrameAlpha(Boomy.starsurgeFrame)

	if Boomy.spells.starsurge.charges > 0 then
		if (Boomy.eclipse.power <= 0 and Boomy.buffs.lunarEmpowerment.charges == 0) or (Boomy.eclipse.power > 0 and Boomy.buffs.solarEmpowerment.charges == 0) or Boomy.spells.starsurge.charges == 3 then -- (eclipse AND 0 empowerment charge) OR 3 starsurge charges
			Boomy:IncreaseFrameAlpha(Boomy.starsurgeFrame)
		end
	end

	-- celestial alignment frame
	Boomy:SetCounter(Boomy.celestialAlignmentFrame, math.ceil(Boomy.spells.celestialAlignment.expirationTime, 0))
	Boomy:DecreaseFrameAlpha(Boomy.celestialAlignmentFrame)

	if Boomy.spells.celestialAlignment.cooldownRemaining == 0 then -- celestial alignement rdy
		Boomy:IncreaseFrameAlpha(Boomy.celestialAlignmentFrame)
	end

	-- incarnation frame
	Boomy:SetCounter(Boomy.incarnationFrame, math.ceil(Boomy.spells.incarnation.expirationTime, 0))
	Boomy:DecreaseFrameAlpha(Boomy.incarnationFrame)

	if Boomy.spells.incarnation.cooldownRemaining == 0 then -- incarnation rdy
		Boomy:IncreaseFrameAlpha(Boomy.incarnationFrame)
	end
end

--
-- CORE
--

function Boomy:GetSpellToCast()
	-- empowered moonkin
	if Boomy.buffs.empoweredMoonkin.expirationTime > 0 then
		return Boomy.buffs.empoweredMoonkin
	end

	-- celestial alignment
	if Boomy.spells.celestialAlignment.expirationTime > 0 then
		if Boomy.eclipse.power <= 0 then
			return Boomy.spells.starfire
		else
			return Boomy.spells.wrath
		end
	end

	-- starfire
	if Boomy.eclipse.direction == "none" or (Boomy.eclipse.direction == "moon" and Boomy.eclipse.power <= 0) or (Boomy.eclipse.direction == "sun" and Boomy.eclipse.power < 0 and Boomy.spells.starfire.castTime <= Boomy.eclipse.remainingTime) or (Boomy.eclipse.direction == "moon" and Boomy.eclipse.power > 0 and Boomy.spells.starfire.castTime >= Boomy.eclipse.remainingTime) then
		return Boomy.spells.starfire
	end

	-- wrath
	return Boomy.spells.wrath
end

function Boomy:Draw(frame, icon)
	frame.texture:SetTexture(icon)
end

function Boomy:IncreaseFrameAlpha(frame)
	frame.texture:SetAlpha(1)
end

function Boomy:DecreaseFrameAlpha(frame)
	frame.texture:SetAlpha(0.25)
end

function Boomy:SetCounter(frame, count)
	if count > 0 then
		frame.fontstring:SetText(count)
		frame.fontstring:Show()
	else
		frame.fontstring:Hide()
	end
end

function Boomy:GetBuff(buffName)
    local _, _, _, count, _, _, expirationTime = UnitAura("player", buffName, nil, "PLAYER|HELPFUL")
    if (expirationTime ~= nil) then
    	expirationTime = expirationTime - GetTime()
        return expirationTime, count
    else
    	return 0, 0
    end
end

function Boomy:GetDebuff(debuffName)
    local _, _, _, _, _, _, expirationTime = UnitAura("target", debuffName, nil, "PLAYER|HARMFUL")
    if (expirationTime ~= nil) then
    	expirationTime = expirationTime - GetTime()
        return expirationTime
    else
    	return 0
    end
end

function Boomy:UnlockFrame(frame)
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetBackdropColor(0, 0, 0, 0.25);
end

function Boomy:LockFrame(frame)
	frame:EnableMouse(false)
	frame:SetMovable(false)
	frame:SetBackdropColor(0, 0, 0, 0);
end

function Boomy:ResetFrame(frame)
	BoomyDB.rp = "CENTER"
	BoomyDB.x = -200
	BoomyDB.y = 200
	frame:SetPoint(BoomyDB.rp, BoomyDB.x, BoomyDB.y)
end

function Boomy:UpdateTalents()
	local _, talent = GetTalentRowSelectionInfo(4)
	if talent == 18579 then
		Boomy.incarnationFrame:Show()
	else
		Boomy.incarnationFrame:Hide()
	end

	local _, talent = GetTalentRowSelectionInfo(7)
	if talent == 21192 then
		Boomy.eclipse.duration = 10
	else
		Boomy.eclipse.duration = 20
	end
end

function Boomy:CreateFrame(parent, scale, x, y, icon)
	local frame = CreateFrame("Frame", nil, parent)
	frame:SetFrameStrata("BACKGROUND")
	frame:SetWidth(BoomyDB.frameUnit * scale)
	frame:SetHeight(BoomyDB.frameUnit * scale)
	frame:SetPoint("CENTER", x, y)
	frame:Show()

	local texture = frame:CreateTexture(nil, "BACKGROUND")
	texture:SetAllPoints(frame)
	if icon then
		texture:SetTexture(icon)
	end
	frame.texture = texture

	local fontString = frame:CreateFontString(nil, "OVERLAY", "GameFontHighlight")
	fontString:SetPoint("CENTER")
	fontString:SetTextHeight(BoomyDB.frameUnit * scale / 2)
	frame.fontstring = fontString

	return frame
end

function Boomy:InitDB()
	BoomyDB = {}
	BoomyDB.frameUnit = 30
	BoomyDB.frames = {}
	BoomyDB.frames.main = {
		rp = "CENTER",
		x = 0,
		y = 0
	}
	BoomyDB.frames.spell = {
		x = 0,
		y = 0
	}
	BoomyDB.frames.moonfire = {
		x = -15,
		y = -45
	}
	BoomyDB.frames.sunfire = {
		x = 15,
		y = -45
	}
	BoomyDB.frames.starfall = {
		x = 45,
		y = 15
	}
	BoomyDB.frames.starsurge = {
		x = 45,
		y = -15
	}
	BoomyDB.frames.celestialAlignment = {
		x = -75,
		y = 15
	}
	BoomyDB.frames.incarnation = {
		x = -75,
		y = -15
	}
end
